package bodybuilder.inspector;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import bodybuilder.util.Config;
import bodybuilder.util.ObjectUtils;

import junit.framework.AssertionFailedError;

/**
 * インスペクター
 */
public abstract class Inspector {

    /////////////////////////////////////////////////////////////////
    // target property

    /**
     * ターゲットクラスのリスト
     */
    private List targets = new ArrayList();

    /**
     * ターゲットクラスのリストを取得する。
     * 
     * @return ターゲットクラスのリスト
     */
    public List getTargets() {
        return targets;
    }

    /**
     * ターゲットクラスを追加する。
     * 
     * @param target ターゲットクラス
     */
    public void addTarget(String target) {
        targets.add(target);
    }

    /////////////////////////////////////////////////////////////////
    // abstract method

    /**
     * 二つのオブジェクトが等しいことを表明する。
     * 
     * @param expected 期待する値
     * @param actual 現実の値
     * @param trace バックトレース
     */
    public abstract void assertEquals(Object expected, Object actual,
            ObjectBackTrace trace);

    /////////////////////////////////////////////////////////////////
    // public method

    /**
     * 二つのオブジェクトが等しいことを表明する。
     * 
     * @param expected 期待する値
     * @param actual 現実の値
     */
    public static void assertObjectEquals(Object expected, Object actual) {
        // バックトレースを生成してオブジェクトを検査。
        ObjectBackTrace trace = new ObjectBackTrace();
        assertObjectEquals(expected, actual, trace);
        trace.clear();
    }

    /**
     * 二つのオブジェクトが等しいことを表明する。
     * 
     * @param expected 期待する値
     * @param actual 現実の値
     * @param trace バックトレース
     */
    public static void assertObjectEquals(Object expected, Object actual,
            ObjectBackTrace trace) {
        trace.append(actual);
        // いづれかがnullではないか検査。
        assertNullEquals(expected, actual, trace);

        // ともにnullの場合は処理を抜ける。
        if (expected == null && actual == null) {
            return;
        }

        // 期待するオブジェクトにマップされたインスペクターを取得。
        Inspector inspector = InspectorMapping.getInspector(expected);

        if (inspector != null) {
            // インスペクターがある場合
            // 実際のオブジェクトの実装クラスリストを取得。
            List actualClassNames = ObjectUtils.getClassNames(actual);
            // インスペクターのターゲットクラスリストを取得。
            List targets = inspector.getTargets();
            boolean isImplement = false;

            // インスペクターのターゲットクラスが実際のオブジェクトに実装されているかチェック。
            for (int i = 0; i < targets.size(); i++) {
                if (actualClassNames.contains(targets.get(i))) {
                    isImplement = true;
                    break;
                }
            }

            // インスペクターのターゲットクラスが実装されていない場合はエラー。
            if (!isImplement) {
                rethrow("unimplement class", inspector.getTargets(), actual
                        .getClass().getName(), trace);
            }

            trace.indent();
            // オブジェクトを検査。
            inspector.assertEquals(expected, actual, trace);
            trace.unindent();
        } else {
            // インスペクターがない場合
            // オブジェクトのクラスを検査。
            assertClassEquals(expected, actual, trace);

            // オブジェクトが異なる場合は例外を投げる。
            if (!expected.equals(actual)) {
                rethrow(expected, actual, trace);
            }
        }
    }

    /////////////////////////////////////////////////////////////////
    // utility method

    /**
     * オブジェクトが等しい(いづれかがnullではない)ことを表明する。
     * 
     * @param expected 期待する値
     * @param actual 実際の値
     * @param trace バックトレース
     */
    protected static void assertNullEquals(Object expected, Object actual,
            ObjectBackTrace trace) {
        // いずれかがnullの場合は例外を投げる。
        if ((expected != null && actual == null)
                || (expected == null && actual != null)) {
            rethrow(expected, actual, trace);
        }
    }

    /**
     * クラスが等しいことを表明する。
     * 
     * @param expected 期待する値
     * @param actual 実際の値
     * @param trace バックトレース
     */
    protected static void assertClassEquals(Object expected, Object actual,
            ObjectBackTrace trace) {
        // クラスを検査しない場合は処理を抜ける。
        if (!Config.isInspectionClass()) {
            return;
        }

        // クラスを取得。
        Class expectedClass = expected.getClass();
        Class actualClass = actual.getClass();

        // クラスが異なる場合は例外を投げる。
        if (!expectedClass.getName().equals(actualClass.getName())) {
            rethrow("classes differ", expectedClass, actualClass, trace);
        }
    }

    /**
     * サイズが等しいことを表明する。
     * 
     * @param expected 期待する値
     * @param actual 実際の値
     * @param trace バックトレース
     */
    protected void assertSizeEquals(int expected, int actual,
            ObjectBackTrace trace) {
        assertSizeEquals("size differ", expected, actual, trace);
    }

    /**
     * サイズが等しいことを表明する。
     * 
     * @param message メッセージ
     * @param expected 期待する値
     * @param actual 実際の値
     * @param trace バックトレース
     */
    protected void assertSizeEquals(String message, int expected, int actual,
            ObjectBackTrace trace) {
        // サイズを検査しない場合は処理を抜ける。
        if (!Config.isInspectionSize()) {
            return;
        }

        // サイズが異なる場合は例外を投げる。
        if (expected != actual) {
            rethrow(message, String.valueOf(expected), String.valueOf(actual),
                    trace);
        }
    }

    /**
     * キーセットが等しいことを表明する。
     * 
     * @param expected 期待する値
     * @param actual 実際の値
     * @param trace バックトレース
     */
    protected void assertKeySetEquals(Set expected, Set actual,
            ObjectBackTrace trace) {
        assertKeySetEquals("key set differ", expected, actual, trace);
    }

    /**
     * キーセットが等しいことを表明する。
     * 
     * @param message メッセージ
     * @param expected 期待する値
     * @param actual 実際の値
     * @param trace バックトレース
     */
    protected void assertKeySetEquals(String message, Set expected, Set actual,
            ObjectBackTrace trace) {
        // イテレータを取得。
        Iterator expectedIte = expected.iterator();
        Iterator actualIte = actual.iterator();

        // 期待するキーセットが実際のキーセットに含まれているか検査。
        while (expectedIte.hasNext()) {
            Object expectedElement = expectedIte.next();

            if (!actual.contains(expectedElement)) {
                rethrow(message, expectedElement, null, trace);
            }
        }

        // 厳密に検査する場合、キーセットが完全に一致するか検査。
        if (Config.isInspectionKeySetStrictly()) {
            while (actualIte.hasNext()) {
                Object actualElement = actualIte.next();

                if (!expected.contains(actualElement)) {
                    rethrow(message, null, actualElement, trace);
                }
            }
        }
    }

    /**
     * 文字列が正規表現にマッチすることを表明する。
     * 
     * @param regex 期待する正規表現
     * @param actual 実際の値
     * @param trace バックトレース
     */
    protected void assertRegexEquals(String regex, String actual,
            ObjectBackTrace trace) {
        assertRegexEquals("not macth regex", regex, actual, trace);
    }

    /**
     * 文字列が正規表現にマッチすることを表明する。
     * 
     * @param message メッセージ
     * @param regex 期待する正規表現
     * @param actual 実際の値
     * @param trace バックトレース
     */
    protected void assertRegexEquals(String message, String regex,
            String actual, ObjectBackTrace trace) {
        // いずれもnullの場合は処理を抜ける。
        if (regex == null && actual == null) {
            return;
        }

        // いずれかがnullの場合は例外を投げる。
        if (actual == null || regex == null) {
            rethrow(message, regex, actual, trace);
        }

        // 文字列が正規表現にマッチしない場合は例外を投げる。
        if (!actual.matches(regex)) {
            rethrow(message, regex, actual, trace);
        }
    }

    /**
     * 例外を投げる。
     * 
     * @param expected 期待する値
     * @param actual 実際の値
     * @param trace バックトレース
     */
    protected static void rethrow(Object expected, Object actual,
            ObjectBackTrace trace) {
        rethrow(null, expected, actual, trace);
    }

    /**
     * 例外を
     * 
     * @param message メッセージ
     * @param expected 期待する値
     * @param actual 実際の値
     * @param trace バックトレース
     */
    protected static void rethrow(String message, Object expected,
            Object actual, ObjectBackTrace trace) {
        StringBuffer buffer = new StringBuffer();

        // メッセージをバッファに追加。
        if (message == null) {
            buffer.append("objects differ");
        } else {
            buffer.append(message);
        }

        // ダンプメッセージを組み立てる。
        buffer.append(" ");
        buffer.append("expected:<");
        buffer.append(expected);
        buffer.append("> but was:<");
        buffer.append(actual);
        buffer.append(">");
        buffer.append(System.getProperty("line.separator"));
        buffer.append(trace.dump());

        // 例外を投げる。
        throw new AssertionFailedError(buffer.toString());
    }

}